/************************************************************************/
/*                                                                      */
/*                      Software UART with FIFO, using T1               */
/*                                                                      */
/*              Author: P. Dannegger                                    */
/*                                                                      */
/************************************************************************/
//			Target: ATtiny44
#include "main.h"
#include "suart.h"

// Hilfs-Macros
#define vu8(x)	(*(volatile uint8_t*)&(x))
#define	OPTR18	__asm__ volatile (""::);		// it helps, but why ?

#define BIT_TIME	(uint16_t)( 1.0 * F_CPU / BAUD + 0.5)
#define ROLLOVER( x, max )	x = ++x >= max ? 0 : x
							// count up and wrap around

#ifdef USE_SUART_TX
#define	TX_HIGH		(1<<COM1A1^1<<COM1A0)
#define	TX_LOW		(TX_HIGH^1<<COM1A0)
#define	TX_OUT		TCCR1A				// use compare output mode

#define SET_TX_HIGH()	TX_OUT = TX_HIGH
#define SET_TX_LOW()	TX_OUT = TX_LOW


static uint8_t stx_buff[STX_SIZE];
static uint8_t stx_in;
static uint8_t stx_out;
static uint8_t stx_data;
static uint8_t stx_state;
#endif

#ifdef USE_SUART_RX
static uint8_t srx_buff[SRX_SIZE];
static uint8_t srx_in;
static uint8_t srx_out;
static uint8_t srx_data;
static uint8_t srx_state;
#endif


void suart_init( void )
{
  OCR1A = (uint8_t)(BIT_TIME - 1);

#ifdef USE_SUART_TX
  SET_TX_HIGH();			// set OC1A high, T1 mode 4
#endif

  TCCR1B = 1<<WGM12^1<<CS10
#ifdef USE_SUART_RX
	^1<<ICNC1			// noise canceler, 1-0 transition,
#endif
	;

#ifdef USE_SUART_TX
  TCCR1C = 1<<FOC1A;
#endif

#ifdef USE_SUART_TX
  stx_state = 0;
  stx_in = 0;
  stx_out = 0;
#endif

#ifdef USE_SUART_RX
  srx_in = 0;
  srx_out = 0;
#endif

#ifdef USE_SUART_TX
  STXD_DDR = 1;				// output enable
#endif

#ifdef USE_SUART_RX
  TIFR1 = 1<<ICF1;			// clear pending interrupt
#endif

  TIMSK1 = 0
#ifdef USE_SUART_RX
	^1<<ICIE1			// enable tx and wait for start
#endif
#ifdef USE_SUART_TX
	^1<<OCIE1A
#endif
	;
}


#ifdef USE_SUART_RX
uint8_t kbhit( void )			// check, if rx buffer not empty
{
  return srx_out ^ vu8(srx_in);
}


uint8_t ugetchar( void )		// wait until byte received
{
  uint8_t data;

  while( !kbhit());		     	// until at least one byte in
  data = srx_buff[srx_out];		// get byte
  ROLLOVER( srx_out, SRX_SIZE );
  return data;
}
#endif


#ifdef USE_SUART_TX
void uputchar( uint8_t c )			// transmit byte
{
  uint8_t i = stx_in;

  ROLLOVER( i, STX_SIZE );
  stx_buff[stx_in] = ~c;		// complement for stop bit after data
  while( i == vu8(stx_out));		// until at least one byte free
					// stx_out modified by interrupt !
  stx_in = i;
}


void uputs_( uint8_t *s )			// transmit string from SRAM
{
  while( *s )
    uputchar( *s++ );
}

/* 20110910 -us */
void uputs_p( const char * pgm_s )
{
  uint8_t c;

  while ( (c = (uint8_t)pgm_read_byte( pgm_s ++)) ) {
    uputchar(c);
  }
}

#endif


/******************************	Interrupts *******************************/

#ifdef USE_SUART_RX
ISR( TIMER1_CAPT_vect )				// start detection
{
  int16_t i = ICR1 - BIT_TIME / 2;		// scan at 0.5 bit time

  OPTR18					// avoid disoptimization
  if( i < 0 )
    i += BIT_TIME;				// wrap around
  OCR1B = i;
  srx_state = 10;
  TIFR1 = 1<<OCF1B;				// clear pending interrupt
  if( SRXD_PIN == 0 )				// still low
    TIMSK1 = 1<<OCIE1A^1<<OCIE1B;		// wait for first bit
}


ISR( TIMER1_COMPB_vect )			// receive data bits
{
  uint8_t i;

  switch( --srx_state ){

    case 9:  if( SRXD_PIN == 0 )		// start bit valid
	       return;
	     break;

    default: i = srx_data >> 1;	      		// LSB first
	     if( SRXD_PIN == 1 )
	       i |= 0x80;			// data bit = 1
	     srx_data = i;
	     return;

    case 0:  if( SRXD_PIN == 1 ){		// stop bit valid
	       i = srx_in;
	       ROLLOVER( i, SRX_SIZE );
	       if( i != srx_out ){		// no buffer overflow
		 srx_buff[srx_in] = srx_data;
		 srx_in = i;			// advance in pointer
	       }
	     }
	     TIFR1 = 1<<ICF1;			// clear pending interrupt
  }
  TIMSK1 = 1<<ICIE1^1<<OCIE1A;			// enable next start
}
#endif


#ifdef USE_SUART_TX
ISR( TIMER1_COMPA_vect )			// transmit data bits
{
  if( stx_state ){
    stx_state--;
    SET_TX_HIGH();

    if( stx_data & 1 )				// lsb first
      SET_TX_LOW();
    stx_data >>= 1;
    return;
  }
  if( stx_in != stx_out ){			// next byte to sent
    stx_data = stx_buff[stx_out];
    ROLLOVER( stx_out, STX_SIZE );
    stx_state = 9;				// 8 data bits + stop
    SET_TX_LOW();				// start bit
  }
}
#endif

