/************************************************************************/
/*                                                                      */
/*                      Interrupt driven buffered UART                  */
/*                                                                      */
/*              Author: Peter Dannegger                                 */
/*                                                                      */
/************************************************************************/

#include <avr/interrupt.h>
#include <avr/io.h>

#include "uart0.h"


#ifdef USE_UART0_RX
static uint8_t rx_buff[RX0_SIZE];
static uint8_t rx_in=0;
static uint8_t rx_out=0;
#endif

#ifdef USE_UART0_TX
static uint8_t tx_buff[TX0_SIZE];
static uint8_t tx_in=0;
static uint8_t tx_out=0;
#endif


#define ROLLOVER( x, max )	x = (++x >= max) ? 0 : x

#ifdef URSEL0		// atMega8
# define URSEL0_SHARE	(1<<URSEL0)	// if UCSR0C shared with UBRR0H
#else
# define URSEL0_SHARE	(0)
#endif

void init_uart0( uint16_t bauddivider, uint8_t databits, uint8_t parity, uint8_t stopbits )
{
  UBRR0H = (uint8_t)(bauddivider >> 8);
  UBRR0L = (uint8_t)bauddivider;	// set baud rate
  UCSR0A = 0;				// no U2X, MPCM

  switch (databits) {
    case 5: UCSR0C = URSEL0_SHARE; break;				// 5-Bit
    case 6: UCSR0C = URSEL0_SHARE | (1<<UCSZ00); break;			// 6-Bit
    case 7: UCSR0C = URSEL0_SHARE | (1<<UCSZ01); break;			// 7-Bit
    case 8: 
    default: UCSR0C = URSEL0_SHARE |(1<<UCSZ01) | (1<<UCSZ00); break;	// 8-Bit
  }

  switch (parity) {
    case UART0_NONE_PARITY: break;
    case UART0_EVEN_PARITY: UCSR0C |= (1<< UPM01); break;
    case UART0_ODD_PARITY : UCSR0C |= (1<< UPM01) | (1<< UPM00); break;
  }

  switch (stopbits) {
    case 0:
    case 1: break;
    case 2:
    default: UCSR0C |= (1<<USBS0); break;
  }

#ifdef USE_UART0_RX
  rx_in = rx_out;			// set buffer empty
#endif

#ifdef USE_UART0_TX
  tx_in = tx_out;
#endif

  UCSR0B = 0 
#ifdef USE_UART0_RX
  	| (1<<RXEN0)	 		// enable RX
	| (1<<RXCIE0)			// enable RX interrupt
#endif

#ifdef USE_UART0_TX
  	| (1<<TXEN0)			// enable TX
#endif
  ;
}


/*
 * UART0 RX
 */

#ifdef USE_UART0_RX
uint8_t ukbhit0( void )
{
  return rx_out ^ vu8(rx_in);		// rx_in modified by interrupt !
}


uint8_t ugetchar0( void )
{
  uint8_t data;

  while( !ukbhit0() );			// until at least one byte in
  data = rx_buff[rx_out];		// get byte
  ROLLOVER( vu8(rx_out), RX0_SIZE );
  URX0_IEN = 1;				// enable RX interrupt
  return data;
}


ISR( USART0_RX_vect )
{
  uint8_t i = rx_in;

  ROLLOVER( i, RX0_SIZE );
  if( i == rx_out ){			// buffer overflow
    URX0_IEN = 0;			// disable RX interrupt
    return;
  }
  rx_buff[rx_in] = UDR0;
  rx_in = i;
}
#endif


/*
 * UART0 TX
 */

#ifdef USE_UART0_TX
ISR( USART0_UDRE_vect )
{
  if( tx_in == tx_out ){		// nothing to sent
    UTX0_IEN = 0;			// disable TX interrupt
    return;
  }
  UDR0 = tx_buff[tx_out];
  ROLLOVER( tx_out, TX0_SIZE );
}


uint8_t utx0_ready( void )
{
  uint8_t i = tx_in;

  ROLLOVER( i, TX0_SIZE );
  return vu8(tx_out) ^ i;		// 0 = busy
}


void uputchar0( uint8_t c )
{
  uint8_t i = tx_in;

  ROLLOVER( i, TX0_SIZE );
  tx_buff[tx_in] = c;
  while( i == vu8(tx_out));		// until at least one byte free
					// tx_out modified by interrupt !
  vu8(tx_in) = i;
  UTX0_IEN = 1;                         // enable TX interrupt
}


void uputs0_( uint8_t *s )
{
  while( *s )
    uputchar0( *s++ );
}

/* 20110312 -us */
void uputs0_p( const char * pgm_s )
{
  uint8_t c;

  while ( (c = (uint8_t)pgm_read_byte( pgm_s ++)) ) {
    uputchar0(c);
  }
}

#endif
